import pygame
from pygame.locals import *
from math import sqrt, ceil, floor
from world import World
from matrix import Matrix
import string
from textrect import render_textrect
import random

##############################
#CONSTANTS
##############################

#Screen
SCREEN_SIZE = (1024, 576)
SCREEN_MARGIN = (10, 10)

#Colors
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
RED = (255, 0, 0)

GRAY1 = (182, 182, 182)
GRAY2 = (155, 155, 155)
GRAY3 = (126, 126, 126)
GRAY4 = (92, 92, 92)
GRAY5 = (72, 72, 72)

GRAYS = [GRAY1, GRAY2, GRAY3, GRAY4, GRAY5]

#Board components
BOARD_SIZE = (7, 6)
AUCTION_BLOCK_SIZE = 4

#Cards
CARD_SIZE = (70, 70)
CARD_MARGIN = (10, 10)
CARD_PADDING = (3, 3)

#Card types
BLANK = 0
COUNTRY = 1
TRANSPORT = 2

#Transport types
PLANE = 0
TRAIN = 1
CAR = 2
SHIP = 3

class Interface(object):
    
    def __init__(self, continent):
        self.board = Board(BOARD_SIZE)
        self.deck = Deck(continent)
        self.auction_block = AuctionBlock()
        self.map = Map(continent)
        self.background = pygame.surface.Surface(SCREEN_SIZE).convert()
        self.background.fill((255, 255, 255)) 
        
        #Set pos of interface elements
        board_pos = (SCREEN_MARGIN[0], SCREEN_MARGIN[1])
        self.board.set_pos(board_pos)
        map_pos = (SCREEN_SIZE[0] - self.map.size[0] - SCREEN_MARGIN[0], 2*SCREEN_MARGIN[1] + CARD_SIZE[1]);
        self.map.set_pos(map_pos)
        auction_block_pos = (map_pos[0] + CARD_MARGIN[0], map_pos[1])
        self.auction_block.set_pos(auction_block_pos)        
        
        #Deal initial cards into auction_block from deck
        self.deck.shuffle()
        self.deck.deal(self.auction_block)
 
    def process(self):
        for event in pygame.event.get():
            #Handle quit
            if event.type == QUIT:
                exit()       
            #Handle mouse clicks
            elif event.type == pygame.MOUSEBUTTONDOWN:
                #Click on auction_block card
                for card in self.auction_block.array:
                    if card.test_pos_on_card(event.pos):
                        if card.selected:
                            card.toggle_select()
                            self.auction_block.selected_card = None
                        else:
                            self.auction_block.select_card(card)
                        return
                #Click on board card
                for card, r, c in self.board.matrix:
                    if card[r][c].test_pos_on_card(event.pos):
                        if self.auction_block.selected_card:
                            old_card = card[r][c]
                            new_card = self.auction_block.selected_card
                            self.auction_block.remove_card(new_card)
                            new_card.set_pos(old_card.pos)
                            self.board.place_card(r, c, new_card)
                            self.auction_block.selected_card = None 
                            new_card.selected = False
                            self.deck.deal(self.auction_block) 
                            self.board.refresh_connections() 
                            if old_card.type != BLANK:    
                                self.deck.discard_stack.append(old_card)                    
        
    def render(self, surface):
        surface.blit(self.background, (0, 0))
        self.map.render(surface)        
        self.board.render(surface)
        self.auction_block.render(surface)

class Map(object):
    def __init__(self, continent):
        self.map_image = pygame.image.load(continent.map).convert()
        self.size = self.map_image.get_size()
        self.pos = (0, 0)
        
    def set_pos(self, pos):
        self.pos = pos    
        
    def render(self, surface):
        surface.blit(self.map_image, (self.pos))

class Deck(object):
    
    def __init__(self, continent):
        self.stack = []
        self.discard_stack = []
        self.continent = continent      
        
        #Add countries  
        for country in continent.countries:
            self.stack.append(CountryCard(country))
        
        #Add ships
        for ocean in continent.oceans:
            for _ in xrange(ocean.number):
                self.stack.append(ShipCard(ocean))
                
        #Add planes
        for color in continent.colors:
            for _ in xrange(2):
                self.stack.append(PlaneCard(color))
                
        #Add cars
        for _ in xrange(5):
            self.stack.append(CarCard())
            
    
    def shuffle(self):
        random.shuffle(self.stack)
            
    def deal(self, auction_block):
        for index, _ in enumerate(auction_block.array):
            old_card = auction_block.array[index]
            try:
                new_card = self.stack.pop(0)
                new_card.set_pos(old_card.pos)
                auction_block.array[index] = new_card
            except IndexError:
                self.merge_discard_stack()
                self.shuffle()
                new_card = self.stack.pop(0)
                new_card.set_pos(old_card.pos)
                auction_block.array[index] = new_card               
            if old_card.type != BLANK:    
                self.discard_stack.append(old_card)
        
    def merge_discard_stack(self):
        for card in self.discard_stack:
            self.stack.append(card)
        self.discard_stack = []
        
class Board(object):
    
    def __init__(self, size):
        self.pos = (0, 0)
        self.matrix_size = size
        self.connections = []        
        
        #Create a matrix to hold the cards, and deal blank cards to start
        self.matrix = Matrix(self.matrix_size[0], self.matrix_size[1])
        for _, r, c in self.matrix:
            blank_card = BlankCard()
            blank_card.set_index([r, c]) 
            self.matrix.setitem(r, c, blank_card)           
        
    def set_pos(self, pos):
        self.pos = pos    
        for card,r,c in self.matrix:
            x = c * (CARD_SIZE[0] + CARD_MARGIN[0]) + self.pos[0]
            y = r * (CARD_SIZE[1] + CARD_MARGIN[1]) + self.pos[1] 
            card[r][c].set_pos((x,y))  
            
    def place_card(self, r, c, card):
        self.matrix.setitem(r, c, card) 
        card.set_index([r, c])       
            
    #Return the card at the given matrix index (r, c)            
    def get_card(self, index):
        return self.matrix.getitem(index[0], index[1])    
    
    #refresh all connections on the board - typically done when a card is moved
    def refresh_connections(self):
        self.connections = []
        for card, r, c in self.matrix:
            current_card = card[r][c]
            try:
                card_right = card[r][c+1]
                if current_card.test_connection(card_right, self):
                    new_connection = Connection(current_card, card_right)
                    self.connections.append(new_connection)
            except:
                pass
            try:
                card_down = card[r+1][c]
                if current_card.test_connection(card_down, self):
                    new_connection = Connection(current_card, card_down)
                    self.connections.append(new_connection)
            except:
                pass       
    
    #Render board components
    def render(self, surface):
        #Render connections
        for connection in self.connections:
            connection.render(surface)            
        #Render cards
        for card,r,c in self.matrix:
            if card[r][c] != 0:
                card[r][c].render(surface)             
    
class AuctionBlock(object):  
    
    def __init__(self):
        self.array = []
        self.pos = (0, 0)
        self.selected_card = None
        for _ in xrange(AUCTION_BLOCK_SIZE):
            self.array.append(BlankCard())

    def set_pos(self, pos):
        self.pos = pos    
        for index, card in enumerate(self.array):
            x = index * (CARD_SIZE[0] + CARD_MARGIN[0]) + self.pos[0]
            y = CARD_MARGIN[1] + self.pos[1]             
            card.set_pos((x, y))
    
    def select_card(self, card):
        for card_i in self.array:
            if card_i == card:
                card_i.selected = True
                self.selected_card = card_i
            else:
                card_i.selected = False
                
    def remove_card(self, card):
        self.array.remove(card)
        self.array.append(BlankCard())
        self.set_pos(self.pos)
        
    def __repr__(self):
        return str(self.array)
        
    def render(self, surface):
        for card in self.array:
            card.render(surface)
    
class Card(object):
    
    def __init__(self, type, text, image, color):  
        self.font = pygame.font.Font("assets/fonts/verdana.TTF", 10)
        self.text_color = BLACK
        self.text_bgcolor = WHITE             
        self.pos = [0, 0]       
        self.type = type     
        self.set_text(text)
        self.color = color
        self.set_image(image)   
        self.selected = False  
        
    def set_text(self, text):
        self.text = text
        self.rendered_text = self.render_text_rec(text)  
        self.text_size = self.rendered_text.get_size()    
        
        self.rendered_price = self.font.render("$" + str(self.price), True, BLACK, WHITE)        

    def set_image(self, image):
        self.image = image
        try:
            image = pygame.image.load(image).convert_alpha()  
            
            #Correct Color
            if self.color:
                image_pixel_array = pygame.PixelArray(image)
                try:
                    image_pixel_array.replace(GRAY1, self.color)
                    image_pixel_array.replace(GRAY2, self.color)
                    image_pixel_array.replace(GRAY3, self.color)
                    image_pixel_array.replace(GRAY4, self.color)
                    image_pixel_array.replace(GRAY5, self.color)
                except pygame.error, e:
                    print e
                    exit()
            
            #Scale image to card size
            height_ratio = float(image.get_height()) / float(image.get_width())
            max_image_height = int(CARD_SIZE[1] - self.text_size[1] - 2*CARD_PADDING[1])
            max_image_width = int(CARD_SIZE[0] - 2*CARD_PADDING[0])
            self.image_size = [0, 0]
            if height_ratio >= float(max_image_height) / float(max_image_width):
                self.image_size[1] = max_image_height
                self.image_size[0] = int(self.image_size[1] / height_ratio) 
            else:
                self.image_size[0] = max_image_width
                self.image_size[1] = int(self.image_size[0] * height_ratio)
                
            #Scale image with scale_factor
            if self.type == COUNTRY:
                scale_factor = self.country_size
            else:
                scale_factor = 0.9                
            self.image_size[0] = int(scale_factor*self.image_size[0])
            self.image_size[1] = int(scale_factor*self.image_size[1])
            
            self.rendered_image = pygame.transform.smoothscale(image, self.image_size)
        except pygame.error, e:
            print e
            exit()
        
    def set_pos(self, pos):
        #Set card pos
        self.pos[0] = pos[0]
        self.pos[1] = pos[1]
        
        #Set text pos
        self.text_pos = [0, 0]
        self.text_pos[0] = self.pos[0] + CARD_PADDING[0]
        self.text_pos[1] = self.pos[1] + CARD_SIZE[1] - CARD_PADDING[1] - self.text_size[1]
        
        #Set price pos
        self.price_pos = [0, 0]
        self.price_pos[0] = self.pos[0] + CARD_PADDING[0]
        self.price_pos[1] = self.pos[1] + CARD_PADDING[1]
        
        #Set img pos
        try:
            self.image_pos = [0, 0]
            max_image_height = int(CARD_SIZE[1] - self.text_size[1] - 2*CARD_PADDING[1])
            self.image_pos[0] = self.get_center()[0] - self.image_size[0]/2
            self.image_pos[1] = self.pos[1] + max_image_height/2 + CARD_PADDING[1] - self.image_size[1]/2
#            self.image_pos[1] = self.get_center()[1] - self.text_size[1] - self.image_size[1]/2 + 2*CARD_PADDING[1]
        except:
            pass     
        
    #Set card's index in the board matrix    
    def set_index(self, index):            
        self.index = index
        
    def toggle_select(self):
        if self.selected:
            self.selected = False
        else:
            self.selected = True
 
    #Return the card's center position
    def get_center(self):
        return (self.pos[0] + CARD_SIZE[0]/2, self.pos[1] + CARD_SIZE[1]/2) 
        
    #Render some text in a wrapped text rectangle that fits within the card    
    def render_text_rec(self, text):
        text_size = self.font.size(text)
        if text_size[0] <= (CARD_SIZE[0] - 2*CARD_PADDING[0]):
            text_lines = 1;
        else:
            text_lines = 2;
        text_rect = pygame.Rect(0, 0, CARD_SIZE[0] - 2*CARD_PADDING[0], 1+text_lines*text_size[1])
        return render_textrect(text, self.font, text_rect, self.text_color, self.text_bgcolor, 1)       
    
    #Test to see if a given position (x,y) is covered by this card, used to check if clicked on card, etc.
    def test_pos_on_card(self, pos):
        if pos[0] > self.pos[0] and pos[0] < self.pos[0] + CARD_SIZE[0]:
            if pos[1] > self.pos[1] and pos[1] < self.pos[1] + CARD_SIZE[1]:
                return True
        return False         
    
    #Test whether a particular card is next to this one on the board
    def test_next_to(self, card):
        if abs(self.index[0] - card.index[0]) == 1 and self.index[1] == card.index[1]:
            return True
        elif abs(self.index[1] - card.index[1]) == 1 and self.index[0] == card.index[0]:
            return True
        else:
            return False     
        
    #Test whether there is a connection between a particular card and this one    
    #Returns True or False
    def test_connection(self, card, board):
        if self.type == BLANK or card.type == BLANK:
            return False
        if self.type == COUNTRY:
            if card.type == COUNTRY:   
                for country_name in self.borders:
                    if country_name == card.country_name:
                        return True
            elif card.type == TRANSPORT:
                if card.transport_type == SHIP:
                    for ocean_name in self.oceans:
                        if ocean_name == card.ocean_name:
                            return True
                elif card.transport_type == PLANE:
                    if self.color == card.color:
                        return True
                elif card.transport_type == CAR:
                    print self
                    print self.index
                    print card
                    print card.index
                    if card.index[1] > self.index[1]:
                        second_card_index = (card.index[0], card.index[1] + 1)
                    else:
                        second_card_index = (card.index[0], card.index[1] - 1)
                    second_card = board.get_card(second_card_index)
                    print second_card
                    print second_card.index
                    for country_name in self.borders:
                        for second_country_name in second_card.borders:
                            if country_name == second_country_name or country_name == second_card.country_name:
                                return True
        elif self.type == TRANSPORT:
            if card.type == TRANSPORT:
                return False
            elif card.type == COUNTRY:
                if self.transport_type == SHIP:
                    for ocean_name in card.oceans:
                        if ocean_name == self.ocean_name:
                            return True        
                elif self.transport_type == PLANE:
                    if self.color == card.color:
                        return True       
                elif self.transport_type == CAR:
                    if self.index[1] > card.index[1]:
                        second_card_index = (self.index[0], self.index[1] + 1)
                    else:
                        second_card_index = (self.index[0], self.index[1] - 1)
                    second_card = board.get_card(second_card_index)
                    for country_name in card.borders:
                        for second_country_name in second_card.borders:
                            if country_name == second_country_name or country_name == second_card.country_name:
                                return True
                      
        return False         
    
    #Test whether this card has any connections, in it's current location
    #Returns True or False
    def test_any_connections(self, board):
        r = self.index[0]
        c = self.index[1]
        try: 
            card_right = board.get_card([r, c+1])
            if self.test_connection(card_right, board):
                return True
        except:
            pass
        try: 
            card_down = board.get_card([r+1, c])
            if self.test_connection(card_down, board):
                return True
        except:
            pass
        try: 
            card_left = board.get_card([r, c-1])
            if self.test_connection(card_left, board):
                return True
        except:
            pass  
        try: 
            card_up = board.get_card([r-1, c])
            if self.test_connection(card_up, board):
                return True
        except:
            pass    
        return False   
    
    #Replace this card with another card
    def replace(self, card):
        card.set_pos(self.pos)
        card.set_index(self.index)
    
    def __repr__(self):
        return str(self.text)    
    
    def render(self, surface):  
        if self.selected:
            if self.image:
                pygame.draw.rect(surface, (0,0,0), Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])))
                surface.blit(self.rendered_image, self.image_pos, None, 0)
        else:
            if self.rendered_image:
                #Background color
                pygame.draw.rect(surface, WHITE, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])))
                #Image
                surface.blit(self.rendered_image, self.image_pos, None, 0)
                #Outline
                pygame.draw.rect(surface, BLACK, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])), 1)
            else:
                pygame.draw.rect(surface, self.color, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])))
        #Text
        surface.blit(self.rendered_text, self.text_pos)
           
        #Price 
        
        try:
            surface.blit(self.rendered_price, self.price_pos)
        except Exception, e:
            print e

    
class BlankCard(Card):
        
    def __init__(self):
        self.price = 0
        # Call the base class constructor
        Card.__init__(self, BLANK, "Blank", "assets/images/blank_card.png", RED)     
    
class CountryCard(Card):
        
    def __init__(self, country):
        self.country_name = country.name
        self.borders = country.borders
        self.oceans = country.oceans
        self.country_size = country.size
        self.price = country.price        
        # Call the base class constructor
        Card.__init__(self, COUNTRY, country.name, country.image, country.color)  
        
class TransportCard(Card):
        
    def __init__(self, transport_type, text, image, color):
        # Call the base class constructor
        Card.__init__(self, TRANSPORT, text, image, color)  
        self.transport_type = transport_type
        
class PlaneCard(TransportCard):
        
    def __init__(self, color):
        plane_image = 'assets/images/transport/plane.png'
        self.price = 100
        # Call the base class constructor
        TransportCard.__init__(self, PLANE, "Airplane", plane_image, color)  
        
class TrainCard(TransportCard):
        
    def __init__(self):
        train_image = 'assets/images/transport/train.png'
        # Call the base class constructor
        TransportCard.__init__(self, TRAIN, "Railroad", train_image, None)   
        
class CarCard(TransportCard):
        
    def __init__(self):
        car_image = 'assets/images/transport/car.png'
        self.price = 200          
        # Call the base class constructor
        TransportCard.__init__(self, CAR, "Car", car_image, None)     
        
class ShipCard(TransportCard):
        
    def __init__(self, ocean):
        ship_image = 'assets/images/transport/ship.png'
        self.ocean_name = ocean.name       
        self.price = len(ocean.borders)*10              
        # Call the base class constructor
        TransportCard.__init__(self, SHIP, ocean.name + " Ship", ship_image, None)                      
    
class Connection(object):    
    
    def __init__(self, card1, card2):
        self.card1 = card1
        self.card2 = card2
        self.point_a = card1.get_center()
        self.point_b = card2.get_center()
        
    def __repr__(self):
        return self.card1.name + "--" + self.card2.name
    
    def render(self, surface):
        pygame.draw.aaline(surface, (0, 0, 0), self.point_a, self.point_b)
    
#Run game 
def main():
    global global_continent
    pygame.init()
    screen = pygame.display.set_mode(SCREEN_SIZE)
    continent = World().europe
    interface = Interface(continent)
        
    #Main game loop    
    while True:  
                            
        interface.process()     
        
        interface.render(screen)
        
        pygame.display.update()

# this runs the main function if this script is called to run.
#  If it is imported as a module, we don't run the main function.
if __name__ == "__main__":
    main()