import pygame
from pygame.locals import *
from math import sqrt, ceil, floor
from world import World
from matrix import Matrix
import string
from textrect import render_textrect
import random
from copy import copy, deepcopy
import operator
from events import *
from event_manager import *
from widgets import *
from card_tree import *
from collections import deque

##############################
#CONSTANTS
##############################

#Screen
SCREEN_SIZE = (1024, 576)
SCREEN_MARGIN = (10, 10)

#Colors
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
RED = (255, 0, 0)
YELLOW = (255, 200, 0)

GRAY1 = (182, 182, 182)
GRAY2 = (155, 155, 155)
GRAY3 = (126, 126, 126)
GRAY4 = (92, 92, 92)
GRAY5 = (72, 72, 72)

GRAYS = [GRAY1, GRAY2, GRAY3, GRAY4, GRAY5]

#Directions
LEFT = -1
RIGHT = 1
UP = 2
DOWN = -2 

#Board components
BOARD_SIZE = (7, 6)
TRIP_LENGTH = 15    #How many cards in a trip

#Cards
CARD_SIZE = (70, 70)
CARD_MARGIN = (10, 10)
CARD_PADDING = (3, 3)

#Card types
BLANK = 0
COUNTRY = 1
TRANSPORT = 2

#Transport types
PLANE = 0
TRAIN = 1
CAR = 2
SHIP = 3

class Interface(object):
    
    def __init__(self, continent, screen, ev_manager):
        self.screen = screen
        self.ev_manager = ev_manager
        ev_manager.register_listener(self)
        self.board = Board(BOARD_SIZE, ev_manager)
        self.deck = Deck(ev_manager, continent)
        self.map = Map(continent)
        self.background = pygame.surface.Surface(SCREEN_SIZE).convert()
        self.background.fill((255, 255, 255)) 
        
        #Set pos of interface elements
        board_pos = (SCREEN_MARGIN[0], SCREEN_MARGIN[1])
        self.board.set_pos(board_pos)
        map_pos = (SCREEN_SIZE[0] - self.map.size[0] - SCREEN_MARGIN[0], 45);
        self.map.set_pos(map_pos)        
        
        #Deal initial cards into board from deck
#        self.deck.shuffle()
#        self.deck, self.board = self.deck.deal_connection_matrix(self.board, (0,0), 10)
#        self.board = self.board.displace_cards(1, [(0, 0)])
#        self.deck.deal_to_blank_squares(self.board)
        
        self.quit_button = ButtonSprite(self.ev_manager, "Quit", (500, 20), None, QuitEvent())
        self.new_board_button = ButtonSprite(self.ev_manager, "New Board", (560, 20), None, DealNewBoardRequest(self.deck, self.board))
 
    def notify(self, event):
        if isinstance(event, TickEvent):          
            self.render(self.screen) 
            pygame.display.update()
        
    def render(self, surface):
        surface.blit(self.background, (0, 0))
        self.map.render(surface)        
        self.board.render(surface)
        self.new_board_button.render(surface)
        self.quit_button.render(surface)

class Map(object):
    def __init__(self, continent):
        self.map_image = pygame.image.load(continent.map).convert()
        self.size = self.map_image.get_size()
        self.pos = (0, 0)
        
    def set_pos(self, pos):
        self.pos = pos    
        
    def render(self, surface):
        surface.blit(self.map_image, (self.pos))

class Deck(object):
    
    def __init__(self, ev_manager, continent = None):
        self.ev_manager = ev_manager     
        self.ev_manager.register_listener(self)   
        self.stack = []
        self.discard_stack = []        
        self.continent = continent  
        if continent != None:  
            self.rebuild(continent)  
            
    def clone(self):
        new_deck = Deck(self.ev_manager)
        for card in self.stack:
            new_card = card.clone()
            new_deck.stack.append(new_card)
        for card in self.discard_stack:
            new_card = card.clone()
            new_deck.discard_stack.append(new_card)
        return new_deck
        
        new_deck.stack = self.stack
        new_deck.discard_stack = self.discard_stack
        return new_deck      
    
    def rebuild(self, continent):
        self.stack = []
        self.discard_stack = []
                
        #Add countries  
        for country in continent.countries:
            for _ in xrange(country.number):
                self.stack.append(CountryCard(country, self.ev_manager))
        
        #Add ships
#        for ocean in continent.oceans:
#            for _ in xrange(ocean.number):
#                self.stack.append(ShipCard(ocean, self.ev_manager))
                
        #Add planes
#        for color in continent.colors:
#            for _ in xrange(1):
#                self.stack.append(PlaneCard(color, self.ev_manager))
                
        #Add cars
#        for _ in xrange(5):
#            self.stack.append(CarCard())        
        
    def shuffle(self):
        random.shuffle(self.stack)
            
    def deal_random(self, board):
        for _, r, c in board.matrix:
            old_card = board.get_card((r, c))
            try:
                new_card = self.stack.pop(0)
                new_card.set_pos(old_card.pos)
                board.place_card(new_card, (r,c))
            except IndexError:
                self.merge_discard_stack()
                self.shuffle()
                new_card = self.stack.pop(0)
                new_card.set_pos(old_card.pos)  
                board.place_card(new_card, (r,c))             
            if old_card.type != BLANK:    
                self.discard_stack.append(old_card)       
#        board.refresh_trip()
        board.refresh_connections()   
    
    #Returns the given board with a connection sequence of the specified number of connections   
    def deal_connection_matrix(self, board, starting_index, number_of_connections):
        number_of_tries = 0
        try_again = True
        while try_again == True:
            number_of_tries += 1
            print 'Connection Matrix Number of tries: ' + str(number_of_tries)          
            if number_of_tries >= 1000:
                raise Exception('Was unable to create connection matrix after ' + str(number_of_tries) + ' tries.' )
                exit()            

            self.merge_board_stack(board)
            self.shuffle    
            current_row = starting_index[0]
            current_col = starting_index[1]
            protected_indexes = []                
            
            #Set starting card
            for card in self.stack:
                if card.type == COUNTRY:
                    self.place_card_on_board(card, board, (current_row, current_col))
                    protected_indexes.append((current_row, current_col))
                    board.start_card = card
                    last_card = card
                    break       
                
            #Create connections
            test_counter = 0
            for i in xrange(number_of_connections - 1):
                test_counter += 1
                
                #Determine which direction to go for connection
                possible_new_indexes = []    
                adjacent_indexes = last_card.get_adjacent_indexes(board)
                for adjacent_index in adjacent_indexes:
                    if adjacent_index not in protected_indexes:
                        possible_new_indexes.append(adjacent_index)
                if len(possible_new_indexes) == 0:      
                    print 'possible new index trouble'         
                    try_again = True
                    break               
                random.shuffle(possible_new_indexes)
                current_row, current_col = possible_new_indexes.pop()     
                
                #Find and place card with connection to last card 
                for card in self.stack:
                    found_connection = False
                    card.set_index((current_row, current_col), board)
                    if i == (number_of_connections - 1) and card.type == TRANSPORT:                   
                        continue
                    if card.type == TRANSPORT and card.transport_type == CAR:   #Special case for cars
                        if last_card.type == COUNTRY:
                            found_connection = True
                            break           
                    if card.test_connection(last_card, board):
                        found_connection = True
                        break
                    
                if found_connection == True:
                    try_again = False
                    self.place_card_on_board(card, board, (current_row, current_col))              
                    protected_indexes.append((current_row, current_col))  
                    last_card = card     
                else: 
                    try_again = True
                    break

        board.refresh_trip()
        board.refresh_connections()
    
    def deal_to_blank_squares(self, board):
        for card, r, c in board.matrix:
            if card[r][c].type == BLANK:
                new_card = self.stack.pop()
                board.place_card_first_time(new_card, (r, c))
        board.refresh_trip()
        board.refresh_connections()          
        
    #displace_by = number of cards to displace each card by  
    #cards with indexes in protected_indexes are not displaced
    #returns False if unsuccessful at displacing cards; can try again
    def displace_cards(self, board, displace_by, protected_indexes):
        original_indexes = {}
        for card, r, c in self.matrix:
            original_indexes[card[r][c]] = card[r][c].index
        print original_indexes
        number_of_tries = 0
        try_again = True
        while try_again == True:   
            
            #Throw exception if too many tries
            number_of_tries += 1
            print 'Number of displacement tries: ' + str(number_of_tries)
            if number_of_tries >= 1000:
                raise Exception('Was unable to create displacement board ' + str(number_of_tries) + ' tries.' )
                exit()        
                
            for card, r, c in self.matrix:
                card[r][c].index = original_indexes[card[r][c]] 
            
            #Set up variables    
            new_protected_indexes = []    
            last_direction = None
                
            for card, r, c in self.matrix:
                if card[r][c].type != BLANK and card[r][c].index not in protected_indexes:      
                    
                    #Determine which direction to go for displacement
                    possible_new_indexes = []    
                    row = card[r][c].index[0]
                    col = card[r][c].index[1]
                    if row != 0 and last_direction != UP:
                        if (row - 1, col) not in new_protected_indexes:
                            possible_new_indexes.append((row - 1, col))
                            last_direction = UP
                    if row != self.matrix.last_row and last_direction != DOWN: 
                        if (row + 1, col) not in new_protected_indexes:
                            possible_new_indexes.append((row + 1, col))
                            last_direction = DOWN
                    if col != 0 and last_direction != LEFT:
                        if (row, col - 1) not in new_protected_indexes:
                            possible_new_indexes.append((row, col - 1))
                            last_direction = LEFT
                    if col != self.matrix.last_col and last_direction != RIGHT: 
                        if (row, col + 1) not in new_protected_indexes:
                            possible_new_indexes.append((row, col + 1))
                            last_direction = RIGHT
                        
#                    print possible_new_indexes   
#                    count = 0 
#                    for index in possible_new_indexes:
#                        count += 1
#                        if index in new_protected_indexes:
#                            print 'yeppers'
#                            possible_new_indexes.remove(index)
#                    print 'Times through: ' + str(count)
#                    print new_protected_indexes        
#                    print possible_new_indexes 
                    
                    if len(possible_new_indexes) == 0:
                        try_again = True
                        break 
                    else:
                        try_again = False             
                        random.shuffle(possible_new_indexes)
                        new_index = possible_new_indexes.pop()
                        self.place_card_first_time(card[r][c], new_index)
                        new_protected_indexes.append(new_index)   
         
#        print new_protected_indexes  
        self.refresh_trip()
        self.refresh_connections()         
        
    def discard(self, card):
        self.stack.remove(card)
        self.discard_stack.append(card)   

    def place_card_on_board(self, card, board, index):
        board.place_card(card, index)
        self.stack.remove(card)
        
    def contains(self, card):
        try:
            return self.stack.index(card)
        except ValueError:
            return False
        
    def merge_discard_stack(self):
        for card in self.discard_stack:
            self.stack.append(card)
        self.discard_stack = []
        
    def merge_board_stack(self, board):
        for card, r, c in board.matrix:
            if card[r][c].type != BLANK:
                self.stack.append(card[r][c])
                board.remove((r, c))
        
    def notify(self, event):    
        #DealNewBoardRequest
#        if isinstance(event, DealNewBoardRequest) and event.deck == self: 
#            self.merge_board_stack(event.board)
#            self.merge_discard_stack()
#            self.shuffle()
#            self.deal_connection_matrix(event.board, (0,0), TRIP_LENGTH)        
#            trip_tree = event.board.build_longest_trip_tree(event.board.start_card)
#            print trip_tree
        
        if isinstance(event, DealNewBoardRequest) and event.deck == self: 
            count = 0
            do_again = True
            while do_again:
                count += 1
                print 'Count: ' + str(count)
                self.merge_board_stack(event.board)
                self.merge_discard_stack()
                self.shuffle()
                self.deal_connection_matrix(event.board, (0,0), TRIP_LENGTH)
                if event.board.longest_trip[0] != event.board.count_cards():
                    do_again = False
#            event.board.displace_cards(1, [(0, 0)])
#            self.deal_to_blank_squares(event.board)                   
        
class Board(object):
    
    def __init__(self, size, ev_manager):
        self.pos = (0, 0)
        self.matrix_size = size
        self.connections = []
        self.selected_card = None 
        self.start_card = None  
        self.longest_trip = None
        self.ev_manager = ev_manager
        
        #Create a matrix to hold the cards, and deal blank cards to start
        self.matrix = Matrix(self.matrix_size[0], self.matrix_size[1])
        self.clear()
        
    def clone(self):
        new_board = Board((self.matrix_size[0], self.matrix_size[1]), self.ev_manager)
        new_board.set_pos(self.pos)
        for card, r, c in self.matrix:
            new_card = card[r][c].clone()
            if card[r][c] == self.start_card:
                new_board.set_start_card(new_card)
            new_board.place_card(new_card, (r, c))
        return new_board
               
     
    def count_cards(self):
        count = 0
        for card, r, c in self.matrix:
            if card[r][c].type != BLANK:
                count += 1
        return count
     
    def clear(self):
        for _, r, c in self.matrix:
            blank_card = BlankCard(self.ev_manager)
            blank_card.set_index((r, c), self) 
            self.matrix.setitem(r, c, blank_card)     
        self.set_pos(self.pos)
        self.connections = []
        
    #Replace card at given index with blank card
    def remove(self, index):
        blank_card = BlankCard(self.ev_manager)
        self.matrix.setitem(index[0], index[1], blank_card) 
        blank_card.set_index((index[0], index[1]), self)     
        self.set_pos(self.pos)
        self.connections = self.refresh_connections()
        
    def set_pos(self, pos):
        self.pos = pos    
        for card,r,c in self.matrix:
            x = c * (CARD_SIZE[0] + CARD_MARGIN[0]) + self.pos[0]
            y = r * (CARD_SIZE[1] + CARD_MARGIN[1]) + self.pos[1] 
            card[r][c].set_pos((x,y))  
    
    def set_start_card(self, card):
        self.start_card = card
            
    def place_card(self, card, index):
        self.matrix.setitem(index[0], index[1], card) 
        card.set_index((index[0], index[1]), self)      
        self.set_pos(self.pos) 
        self.refresh_connections()
        
    def place_card_first_time(self, card, index):
        card.possible_indexes = []
        self.place_card(card, index)
        card.possible_indexes = list(card.get_adjacent_indexes(self))
        card.possible_indexes.append(index)
            
    #Return the card at the given matrix index (r, c)            
    def get_card(self, index):
        if index[0] < 0 or index[1] < 0:
            raise Exception('Cannot get card from a negative index')
        return self.matrix.getitem(index[0], index[1])            
    
    def test_index_on_board(self, index):
        if index[0] < 0 or index[0] > self.matrix_size[0] - 1:
            return False
        elif index[1] < 0 or index[1] > self.matrix_size[1] - 1:
            return False
        else:
            return True
   
    def clear_selection(self):
        if self.selected_card:
            self.selected_card.selected = False
            self.selected_card = None
            for card, r, c in self.matrix:
                card[r][c].possible_swap = False
   
    def set_possible_swaps(self):
        possible_indexes = self.selected_card.possible_indexes
        for index in possible_indexes:
            card = self.get_card(index)
            card.possible_swap = True
            
    def select_card(self, card):
        self.clear_selection()
        card.selected = True
        self.selected_card = card
        self.set_possible_swaps()
        
   
    def try_swap(self, card1, card2):
        #Change card information for testing
        card1_index_old = card1.index
        card2_index_old = card2.index
        card1.index = card2_index_old
        card2.index = card1_index_old    
        
        #Test for new connections after the swap
        any_connections = card1.get_connections(self) or card2.get_connections(self)
        
        #Change card info back
        card1.index = card1_index_old
        card2.index = card2_index_old  
        
        #If there are connections after the swap, then swap, else don't
        if any_connections:       
            self.swap(card1, card2) 
        else: 
            return False        
        
    def swap(self, card1, card2):
        #change card information
        card1_index_old = card1.index
        card2_index_old = card2.index
        card1.index = card2_index_old
        card2.index = card1_index_old    
        
        card1_pos_old_0 = card1.pos[0]
        card1_pos_old_1 = card1.pos[1]
        card2_pos_old_0 = card2.pos[0]
        card2_pos_old_1 = card2.pos[1]
        card1.set_pos([card2_pos_old_0, card2_pos_old_1])
        card2.set_pos([card1_pos_old_0, card1_pos_old_1])
          
        card1.selected = False
        card2.selected = False
        
        #change board information
        self.selected_card = None 
        self.matrix.setitem(card1_index_old[0], card1_index_old[1], card2)
        self.matrix.setitem(card2_index_old[0], card2_index_old[1], card1)
        self.refresh_trip()        
        self.refresh_connections()     
        
    #displace_by = number of cards to displace each card by  
    #cards with indexes in protected_indexes are not displaced
    #returns False if unsuccessful at displacing cards; can try again
    def displace_cards(self, displace_by, protected_indexes):
        original_setup = {}
        for card, r, c in self.matrix:
            if card[r][c].type != BLANK:
                original_setup[card[r][c]] = card[r][c].index
        #Sort the cards by matrix index, so that they can be displaced more systematically
        #This method attained from http://stackoverflow.com/questions/613183/python-sort-a-dictionary-by-value
        original_setup = sorted(original_setup.items(), key=operator.itemgetter(1))
        
        number_of_tries = 0
        try_again = True
        while try_again == True:   
            #Throw exception if too many tries
            number_of_tries += 1
            print 'Number of displacement tries: ' + str(number_of_tries)
            if number_of_tries >= 1000:
                raise Exception('Was unable to create displacement board ' + str(number_of_tries) + ' tries.' )
                exit()        
            
            self.clear()    
            
            #Set up variables    
            new_protected_indexes = list(protected_indexes)    
            last_direction = None

            for card, original_index in original_setup:
                possible_new_indexes = []
                if original_index in protected_indexes:
                    self.place_card(card, original_index)
                    try_again = False
                else:      
                    #Determine which direction to go for displacement
                    row = original_index[0]
                    col = original_index[1]
                    if row != 0 and last_direction != UP and last_direction != DOWN:
                        if (row - 1, col) not in new_protected_indexes:
                            possible_new_indexes.append((row - 1, col))
                            last_direction = UP
                    if row != self.matrix.last_row and last_direction != DOWN and last_direction != UP: 
                        if (row + 1, col) not in new_protected_indexes:
                            possible_new_indexes.append((row + 1, col))
                            last_direction = DOWN
                    if col != 0 and last_direction != LEFT and last_direction != RIGHT:
                        if (row, col - 1) not in new_protected_indexes:
                            possible_new_indexes.append((row, col - 1))
                            last_direction = LEFT
                    if col != self.matrix.last_col and last_direction != RIGHT and last_direction != LEFT: 
                        if (row, col + 1) not in new_protected_indexes:
                            possible_new_indexes.append((row, col + 1))
                            last_direction = RIGHT
                    
                    if len(possible_new_indexes) == 0:
                        try_again = True
                        break 
                    else:
                        try_again = False             
                        random.shuffle(possible_new_indexes)
                        new_index = possible_new_indexes.pop()
                        self.place_card(card, new_index)
                        new_protected_indexes.append(new_index)   
                    
        self.refresh_trip()
        self.refresh_connections()               
     
    def refresh_trip(self):
        for card, r, c in self.matrix:
            card[r][c].in_trip = False
            card[r][c].in_longest_trip = False
        self.start_card.add_to_trip(self)    
        self.longest_trip = self.get_longest_trip()
        if self.longest_trip:
            for card, r, c in self.matrix:
                if card[r][c] in self.longest_trip[1]:
                    card[r][c].in_longest_trip = True
        
    
    def get_longest_trip(self):
        return self.start_card.get_longest_trip(self, list())    
    
    def build_longest_trip_tree(self, card):
        start_node = CardTreeNode(card)
        tree = CardTree(start_node)
        
        queue = deque()
        queue.append(start_node)
        
        while len(queue) > 0:
            current_node = queue.popleft()
            print 'Current node:' + str(current_node)
            current_card = current_node.card
            if current_card.get_connections(self):
                for child_card in current_node.card.get_connections(self):
                    child_node = CardTreeNode(child_card, current_node)
                    print 'Child node: ' + str(child_node)
                    if child_node.test_ancestor(child_node):
                        print 'yeppers'
                    if not child_node.test_ancestor(child_node):
                        tree.insert_node(child_node, current_node)
                        queue.append(child_node)
                    else:
                        child_node.clear()
                        
        return tree
            
    
    #refresh all connections on the board - typically done when a card is moved
    #IMPORTANT: refresh_trip needs to be called first for connections' in_trip property to be set correctly
    def refresh_connections(self):
        self.connections = []
        for card, r, c in self.matrix:
            current_card = card[r][c]
            try:
                card_right = card[r][c+1]
                if current_card.test_connection(card_right, self):
                    new_connection = Connection(current_card, card_right)
                    self.connections.append(new_connection)
            except:
                pass
            try:
                card_down = card[r+1][c]
                if current_card.test_connection(card_down, self):
                    new_connection = Connection(current_card, card_down)
                    self.connections.append(new_connection)
            except:
                pass  
    
    #Render board components
    def render(self, surface):
        #Render connections
        for connection in self.connections:
            if connection.in_trip:
                connection.render(surface)            
        #Render cards
        for card,r,c in self.matrix:
            if card[r][c] != 0:
                card[r][c].render(surface)             
    
class Card(object):
    
    def __init__(self, type, text, image, color, ev_manager):  
        self.price = 0
        self.font = pygame.font.Font("assets/fonts/verdana.TTF", 10)
        self.text_color = BLACK
        self.text_bgcolor = WHITE             
        self.pos = [0, 0]       
        self.type = type     
        self.set_text(text)
        self.color = color
        self.set_image(image)   
        self.selected = False  
        self.possible_swap = False
        self.in_trip = False
        self.in_longest_trip = False
        self.possible_indexes = ()
        self.ev_manager = ev_manager
        self.ev_manager.register_listener(self)

    def clone(self):
        if self.type == BLANK:
            return BlankCard(self.ev_manager)
        elif self.type == COUNTRY:
            return CountryCard(self.country, self.ev_manager)
        elif self.type == TRANSPORT:
            if self.transport_type == PLANE:
                return PlaneCard(self.color, self.ev_manager)
            elif self.transport_type == TRAIN:
                return TrainCard(self.ev_manager)
            elif self.transport_type == CAR:
                return CarCard(self.ev_manager)    
            elif self.transport_type == SHIP:
                return ShipCard(self.ocean, self.ev_manager)                    
        
    def set_text(self, text):
        self.text = text
        self.rendered_text = self.render_text_rec(text)  
        self.text_size = self.rendered_text.get_size()    
        
        self.rendered_price = self.font.render("$" + str(self.price), True, BLACK, WHITE)        

    def set_image(self, image):
        self.image = image
        try:
            image = pygame.image.load(image).convert_alpha()  
            
            #Correct Color
            if self.color:
                image_pixel_array = pygame.PixelArray(image)
                try:
                    image_pixel_array.replace(GRAY1, self.color)
                    image_pixel_array.replace(GRAY2, self.color)
                    image_pixel_array.replace(GRAY3, self.color)
                    image_pixel_array.replace(GRAY4, self.color)
                    image_pixel_array.replace(GRAY5, self.color)
                except pygame.error, e:
                    print e
                    exit()
            
            #Scale image to card size
            height_ratio = float(image.get_height()) / float(image.get_width())
            max_image_height = int(CARD_SIZE[1] - self.text_size[1] - 2*CARD_PADDING[1])
            max_image_width = int(CARD_SIZE[0] - 2*CARD_PADDING[0])
            self.image_size = [0, 0]
            if height_ratio >= float(max_image_height) / float(max_image_width):
                self.image_size[1] = max_image_height
                self.image_size[0] = int(self.image_size[1] / height_ratio) 
            else:
                self.image_size[0] = max_image_width
                self.image_size[1] = int(self.image_size[0] * height_ratio)
                
            #Scale image with scale_factor
            if self.type == COUNTRY:
                scale_factor = self.country_size
            else:
                scale_factor = 0.9                
            self.image_size[0] = int(scale_factor*self.image_size[0])
            self.image_size[1] = int(scale_factor*self.image_size[1])
            
            self.rendered_image = pygame.transform.smoothscale(image, self.image_size)
        except pygame.error, e:
            print e
            exit()
        
    def set_pos(self, pos):
        #Set card pos
        self.pos[0] = pos[0]
        self.pos[1] = pos[1]
        
        #Set text pos
        self.text_pos = [0, 0]
        self.text_pos[0] = self.pos[0] + CARD_PADDING[0]
        self.text_pos[1] = self.pos[1] + CARD_SIZE[1] - CARD_PADDING[1] - self.text_size[1]
        
        #Set price pos
        self.price_pos = [0, 0]
        self.price_pos[0] = self.pos[0] + CARD_PADDING[0]
        self.price_pos[1] = self.pos[1] + CARD_PADDING[1]
        
        #Set img pos
        try:
            self.image_pos = [0, 0]
            max_image_height = int(CARD_SIZE[1] - self.text_size[1] - 2*CARD_PADDING[1])
            self.image_pos[0] = self.get_center()[0] - self.image_size[0]/2
            self.image_pos[1] = self.pos[1] + max_image_height/2 + CARD_PADDING[1] - self.image_size[1]/2
#            self.image_pos[1] = self.get_center()[1] - self.text_size[1] - self.image_size[1]/2 + 2*CARD_PADDING[1]
        except:
            pass     
        
    #Set card's index in the board matrix    
    def set_index(self, index, board):
        self.board = board            
        self.index = index
        
    def toggle_select(self):
        if self.selected:
            self.selected = False
        else:
            self.selected = True
 
    #Return the card's center position
    def get_center(self):
        return (self.pos[0] + CARD_SIZE[0]/2, self.pos[1] + CARD_SIZE[1]/2) 
        
    #Render some text in a wrapped text rectangle that fits within the card    
    def render_text_rec(self, text):
        text_size = self.font.size(text)
        if text_size[0] <= (CARD_SIZE[0] - 2*CARD_PADDING[0]):
            text_lines = 1;
        else:
            text_lines = 2;
        text_rect = pygame.Rect(0, 0, CARD_SIZE[0] - 2*CARD_PADDING[0], 1+text_lines*text_size[1])
        return render_textrect(text, self.font, text_rect, self.text_color, self.text_bgcolor, 1)       
    
    #Test to see if a given position (x,y) is covered by this card, used to check if clicked on card, etc.
    def test_pos_on_card(self, pos):
        if pos[0] > self.pos[0] and pos[0] < self.pos[0] + CARD_SIZE[0]:
            if pos[1] > self.pos[1] and pos[1] < self.pos[1] + CARD_SIZE[1]:
                return True
        return False         
    
    #Test whether a particular card is next to this one on the board
    def test_next_to(self, card):
        if abs(self.index[0] - card.index[0]) == 1 and self.index[1] == card.index[1]:
            return True
        elif abs(self.index[1] - card.index[1]) == 1 and self.index[0] == card.index[0]:
            return True
        else:
            return False  
    
    def add_to_trip(self, board):
        self.in_trip = True
        if self.get_connections(board):
            for card in self.get_connections(board):
                if card.in_trip == False:
                    card.add_to_trip(board)  
  
    def get_longest_trip(self, board, processed_connections = list()):
        possible_trips = dict()
        if self.get_connections(board):
            for i, card in enumerate(self.get_connections(board)):
                if card.in_trip == True and (self, card) not in processed_connections:
                    processed_connections.append((self, card))
                    processed_connections.append((card, self))
                    possible_trips[i] = card.get_longest_trip(board, processed_connections)
            if possible_trips:    
                longest_trip = max(possible_trips.iteritems(), key=operator.itemgetter(1))[1]
                card_list = longest_trip[1]
                longest_trip = longest_trip[0]
                card_list.append(self)
                return [longest_trip + 1, card_list]
            else:
                card_list = []
                card_list.append(self)
                return [1, card_list]
        else:
            return None
                                 
    def get_adjacent_indexes(self, board):
        row = self.index[0]
        col = self.index[1]
        adjacent_indexes = []
        if row != 0 and (row - 1, col):
            adjacent_indexes.append((row - 1, col))
        if row != board.matrix.last_row: 
            adjacent_indexes.append((row + 1, col))
        if col != 0:
            adjacent_indexes.append((row, col - 1))
        if col != board.matrix.last_col: 
            adjacent_indexes.append((row, col + 1))
        return adjacent_indexes
                
    #Test whether there is a connection between a particular card and this one    
    #Returns True or False
    def test_connection(self, card, board):
        if self.type == BLANK or card.type == BLANK:
            return False
        if self.type == COUNTRY:
            if card.type == COUNTRY:   
                for country_name in self.borders:
                    if country_name == card.country_name:
                        return True
            elif card.type == TRANSPORT:
                if card.transport_type == SHIP:
                    for ocean_name in self.oceans:
                        if ocean_name == card.ocean_name:
                            return True
                elif card.transport_type == PLANE:
                    if self.color == card.color:
                        return True
                elif card.transport_type == CAR:
                    second_card_indexes = []
                    if card.index[1] > self.index[1]:
                        second_card_indexes.append((card.index[0], card.index[1] + 1))
                        second_card_indexes.append((card.index[0] + 1, card.index[1]))
                        second_card_indexes.append((card.index[0] - 1, card.index[1]))
                    elif card.index[1] < self.index[1]:
                        second_card_indexes.append((card.index[0], card.index[1] - 1))
                        second_card_indexes.append((card.index[0] + 1, card.index[1]))
                        second_card_indexes.append((card.index[0] - 1, card.index[1]))
                    elif card.index[0] > self.index[0]:
                        second_card_indexes.append((card.index[0] + 1, card.index[1]))
                        second_card_indexes.append((card.index[0], card.index[1] + 1))
                        second_card_indexes.append((card.index[0], card.index[1] - 1))
                    else:
                        second_card_indexes.append((card.index[0] + -1, card.index[1]))
                        second_card_indexes.append((card.index[0], card.index[1] + 1))
                        second_card_indexes.append((card.index[0], card.index[1] - 1))
                    for second_card_index in second_card_indexes:
                        if board.test_index_on_board(second_card_index):
                            second_card = board.get_card(second_card_index)
                            if second_card.type == COUNTRY:
                                for country_name in self.borders:
                                    for second_country_name in second_card.borders:
                                        if country_name == second_country_name or country_name == second_card.country_name:
                                            return True
        elif self.type == TRANSPORT:
            if card.type == TRANSPORT:
                return False
            elif card.type == COUNTRY:
                if self.transport_type == SHIP:
                    for ocean_name in card.oceans:
                        if ocean_name == self.ocean_name:
                            return True        
                elif self.transport_type == PLANE:
                    if self.color == card.color:
                        return True       
                elif self.transport_type == CAR:
                    second_card_indexes = []
                    if self.index[1] > card.index[1]:
                        second_card_indexes.append((self.index[0], self.index[1] + 1))
                        second_card_indexes.append((self.index[0] + 1, self.index[1]))
                        second_card_indexes.append((self.index[0] - 1, self.index[1]))
                    elif self.index[1] < card.index[1]:
                        second_card_indexes.append((self.index[0], self.index[1] - 1))
                        second_card_indexes.append((self.index[0] + 1, self.index[1]))
                        second_card_indexes.append((self.index[0] - 1, self.index[1]))
                    elif self.index[0] > card.index[0]:
                        second_card_indexes.append((self.index[0] + 1, self.index[1]))
                        second_card_indexes.append((self.index[0], self.index[1] + 1))
                        second_card_indexes.append((self.index[0], self.index[1] - 1))
                    else:
                        second_card_indexes.append((self.index[0] + -1, self.index[1]))
                        second_card_indexes.append((self.index[0], self.index[1] + 1))
                        second_card_indexes.append((self.index[0], self.index[1] - 1))
                    for second_card_index in second_card_indexes:
                        if board.test_index_on_board(second_card_index):
                            second_card = board.get_card(second_card_index)
                            if second_card.type == COUNTRY:
                                for country_name in card.borders:
                                    for second_country_name in second_card.borders:
                                        if country_name == second_country_name or country_name == second_card.country_name:
                                            return True
                      
        return False         
    
    #Get all cards with immediate connections to this card
    #Returns a list of cards, or None if no connections
    def get_connections(self, board):
        r = self.index[0]
        c = self.index[1]
        connections = []
        try: 
            card_right = board.get_card([r, c+1])
            if self.test_connection(card_right, board):
                connections.append(card_right)
        except:
            pass
        try: 
            card_down = board.get_card([r+1, c])
            if self.test_connection(card_down, board):
                connections.append(card_down)
        except:
            pass
        try: 
            card_left = board.get_card([r, c-1])
            if self.test_connection(card_left, board):
                connections.append(card_left)
        except:
            pass  
        try: 
            card_up = board.get_card([r-1, c])
            if self.test_connection(card_up, board):
                connections.append(card_up)
        except:
            pass    
        if len(connections) > 0:
            return connections   
        else:
            return None
    
    #Replace this card with another card
    def replace(self, card):
        card.set_pos(self.pos)
        card.set_index(self.index, self.board)
    
    def __repr__(self):
        return str(self.text)    
    
    def notify(self, event):
        #MouseButton Event
        if isinstance(event, MouseButtonLeftEvent):
            if self.test_pos_on_card(event.pos):
                if self.board.selected_card:
                    if self.selected:
                        self.board.clear_selection()
                    elif self.possible_swap:
                        self.board.swap(self, self.board.selected_card)
                    elif not self.selected:
                        self.board.select_card(self)
                else:
                    self.board.select_card(self)                   
    
    def render(self, surface):  
        if self.selected:
            if self.image:
                pygame.draw.rect(surface, (0,0,0), Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])))
                surface.blit(self.rendered_image, self.image_pos, None, 0)
        else:
            if self.rendered_image:
                #Background color
                if self.in_longest_trip:
                    pygame.draw.rect(surface, RED, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])))
                else:
                    pygame.draw.rect(surface, WHITE, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])))
                #Image
                surface.blit(self.rendered_image, self.image_pos, None, 0)
                #Outline
                if self.possible_swap:
                    pygame.draw.rect(surface, YELLOW, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])), 2)
                else:
                    pygame.draw.rect(surface, BLACK, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])), 1)
            else:
                pygame.draw.rect(surface, self.color, Rect(self.pos, (CARD_SIZE[0], CARD_SIZE[1])))
        #Text
        surface.blit(self.rendered_text, self.text_pos)
           
        #Price 
        
        try:
            surface.blit(self.rendered_price, self.price_pos)
        except Exception, e:
            print e

    
class BlankCard(Card):
        
    def __init__(self, ev_manager):
        self.price = 0
        self.borders = None
        # Call the base class constructor
        Card.__init__(self, BLANK, "Blank", "assets/images/blank_card.png", RED, ev_manager)     
    
class CountryCard(Card):
        
    def __init__(self, country, ev_manager):
        self.country = country
        self.country_name = country.name
        self.borders = country.borders
        self.oceans = country.oceans
        self.country_size = country.size
        self.price = country.price        
        # Call the base class constructor
        Card.__init__(self, COUNTRY, country.name, country.image, country.color, ev_manager)  
        
class TransportCard(Card):
        
    def __init__(self, transport_type, text, image, color, ev_manager):
        # Call the base class constructor
        Card.__init__(self, TRANSPORT, text, image, color, ev_manager)  
        self.transport_type = transport_type
        
class PlaneCard(TransportCard):
        
    def __init__(self, color, ev_manager):
        plane_image = 'assets/images/transport/plane.png'
        self.price = 100
        # Call the base class constructor
        TransportCard.__init__(self, PLANE, "Airplane", plane_image, color, ev_manager)    
        
class TrainCard(TransportCard):
        
    def __init__(self, ev_manager):
        train_image = 'assets/images/transport/train.png'
        # Call the base class constructor
        TransportCard.__init__(self, TRAIN, "Railroad", train_image, None, ev_manager)   
        
class CarCard(TransportCard):
        
    def __init__(self, ev_manager):
        car_image = 'assets/images/transport/car.png'
        self.price = 200          
        # Call the base class constructor
        TransportCard.__init__(self, CAR, "Car", car_image, None, ev_manager)     
        
class ShipCard(TransportCard):
        
    def __init__(self, ocean, ev_manager):
        ship_image = 'assets/images/transport/ship.png'
        self.ocean = ocean
        self.ocean_name = ocean.name       
        self.price = len(ocean.borders)*10              
        # Call the base class constructor
        TransportCard.__init__(self, SHIP, ocean.name + " Ship", ship_image, None, ev_manager)                      
    
class Connection(object):    
    
    def __init__(self, card1, card2):
        self.card1 = card1
        self.card2 = card2
        self.point_a = card1.get_center()
        self.point_b = card2.get_center()
        if self.card1.in_trip or self.card2.in_trip:
            self.in_trip = True
        else:
            self.in_trip = False
          
    def __repr__(self):
        return self.card1.name + "--" + self.card2.name
    
    def render(self, surface):
        pygame.draw.aaline(surface, (0, 0, 0), self.point_a, self.point_b)
    
#Run game 
def main():
    pygame.init()
    screen = pygame.display.set_mode(SCREEN_SIZE)
    continent = World().europe
    
    
    ev_manager = EventManager()
    spinner = CPUSpinnerController(ev_manager)
    interface = Interface(continent, screen, ev_manager)    
    event_controller = PyGameEventController(ev_manager)
    
    spinner.run()
    

# this runs the main function if this script is called to run.
#  If it is imported as a module, we don't run the main function.
if __name__ == "__main__":
    main()